#include <Servo.h>
#include <AccelStepper.h>
#include <Stepper.h>
#include <Wire.h>
#include "Adafruit_TCS34725.h"
#include "Pencils.h"
#include "Color.h"

#include <Arduino_Helpers.h>
#include <AH/STL/algorithm>
#include <AH/STL/iterator>

#define PI 3.1415926535897932384626433832795

//--------------Grabber Servo control-------------
Servo grabberServo;                     //library object
Servo folderServo;                      //library object      
#define fullyOpenAngle 115              //maximum opening angle, equals to the sensor down position
#define openAngle 80                    //angle number for open position before grabbing a pencil
#define closedAngle -2                   
#define defaultGrabberPosition 25       //angle for default resting position
#define grabberServoPin 10              //pin

//--------------Folder Servo control-------------
#define defaultFolderPosition 0          //Position Angles
#define downFolderPosition 82 
#define liftedFolderPosition 65           
#define folderServoPin 9                 //pin

//--------------moveStepper Control-------------
#define FULLSTEP 4
#define stepsPerRevolution 2048 // change this to fit the number of steps per revolution of used motor
const int revolution = stepsPerRevolution * 2;
AccelStepper moveStepper(FULLSTEP, 5, 3, 4, 2); // initialize the moveStepper library on pins 2 through 5:
int stepCount = 0;  // number of steps the motor has taken

//--------------Prototyping inputs Control-------------
#define buttonPin 7     //test button to launch a function
bool buttonState;

//--------------Sensor Control-------------
#define calibrationEnabled true
#define sensorSCL     A5
#define sensorSDA     A4
#define averagingCount 3    //number of sensing iterations to average out 
#define calibrationDuration 5    //calibration duration in seconds
/* Initialise the sensor object with specific integration time and gain values */
Adafruit_TCS34725 tcs = Adafruit_TCS34725(TCS34725_INTEGRATIONTIME_154MS, TCS34725_GAIN_1X); 
uint16_t redLow = 2000;    //Default values for the sensor's calibration. Overriden when calibrating
uint16_t redHigh = 0;
uint16_t greenLow = 2000;
uint16_t greenHigh = 0;
uint16_t blueLow = 2000;
uint16_t blueHigh = 0;

//--------------Pencil Variables-------------
#define penCount 4          //not dynamic array sizing, need to define count before
static Pencil pencils[penCount];  //Array of all the pencils on the table. Empty before exploration
#define exploreIncrement 4

//--------------------------------------------------------------------------------------------------------------------------------------

void setup() {
  Serial.begin(115200);
  Serial.println("  ");
  Serial.println("ENTERING SETUP");
  pinMode(LED_BUILTIN, OUTPUT);
  pinMode(buttonPin, INPUT);

  //Initializing Servo Motors
    grabberServo.write(defaultGrabberPosition);
    grabberServo.attach(grabberServoPin);  // attaches the servo object to corresponding pin
    delay(300);
    folderServo.write(defaultFolderPosition);
    folderServo.attach(folderServoPin);
    delay(3000);
  //folderServo.write(0);  

  
  //Initializing Step Motors
    moveStepper.setMaxSpeed(1000.0);   // set the maximum speed
    moveStepper.setAcceleration(1000.0); // set acceleration
    moveStepper.setSpeed(300);         // set initial speed
    moveStepper.setCurrentPosition(0);  //resets current position

  
  //Initializing Sensor
    bool found = false;
    while (!found) {
      if (tcs.begin()) {
        Serial.println("sensor detected");
        found = true;
      } else {
        Serial.println("No TCS34725 found ... check your connections");
        //while (1);
        delay(400);
      }
    }

  //Sensor calibration. 
    if (calibrationEnabled) {
        sensorCalibration();   
    }

}

//--------------------------------------------------------------------------------------------------------------------------------------

void loop() {

  bool buttonState = digitalRead(buttonPin);
  moveStepper.disableOutputs();

  if (buttonState) {

    moveStepper.enableOutputs();
    
    /*
    Grab();
    delay(400);
    Place();
    */
    
    Explore(250);           //spans the table over 250mm to find all pencils
    sortPencils();          //sorts the list of pencils just created
    for(int p = 0; p < penCount; p++){  //moves all pencils to their sorted position
      MovePencil(pencils[p]);
    }    
  }
}

//------------------------Actions-------------------------------------

//This will let us scan the pencil's position over a 'span' distance in mm
void Explore(unsigned int span) {
  Serial.println("STARTING EXPLORATION");
  //Set color threshold by mesuring background noise
  moveStepper.runToNewPosition(-distanceToStep(10));     //moves away from default position
  Serial.println("Sensing Background 1");
  static Color bgColour1 = senseColour(!calibrationEnabled, true, false);    //stores 1st background color
  moveStepper.runToNewPosition(distanceToStep(10));
  Serial.println("Sensing Background 2");
  static Color bgColour2 = senseColour(!calibrationEnabled, true, false);
  moveStepper.runToNewPosition(0);     //back home
  float threshold = euclidianDistance(bgColour1, bgColour2) * 1.5 + 0.16;   //sets color sensing threshold with added percentage
  Serial.print("Threshold = "); Serial.println(threshold);
  moveStepper.runToNewPosition(-distanceToStep(30));

  //Initializing Step Motors
    moveStepper.setCurrentPosition(0);  //resets current position
    

  int index = 0;                                       //keeps track of how many pencils found
  int distance = distanceToStep(exploreIncrement);
  int nextDistance = distance;
  Serial.print(nextDistance);

  for (int d = 0; d <= span; d+exploreIncrement){

    moveStepper.move(-nextDistance); //move to target position
    moveStepper.setSpeed(200);         // set initial speed
    while (moveStepper.distanceToGo() != 0) {
          moveStepper.run(); }  

    Color currentColour = senseColour(!calibrationEnabled, true, false);
    float diff = euclidianDistance(bgColour1, currentColour);
    Serial.print("Color diff = "); Serial.println(diff);
    
    if (diff > threshold) {    //value needs to be adapted, rn IDK what range to expect
      //moveStepper.run();    //adds one step to get better view of the pencil
      Serial.print("PEN FOUND at ");
      int position = moveStepper.currentPosition();
      Serial.println(position);
      Pencil penFound(currentColour, position);
      pencils[index] = penFound;
      Serial.println("Hue = "); Serial.print(penFound.hue);
      index++;
      nextDistance = distance * 3;
    }
    else { nextDistance = distance; }
    
    if(index == penCount){ 
        Serial.println("PEN COUNT REACHED");
        break; 
    }
  }
}

void MovePencil(Pencil &p) {
  Serial.println("MOVE PENCIL");
  GotoPencil(p); 
  Grab();
  delay(200);
  Serial.print("PLACING PENCIL at ");
  int pos = p.sortedPosition;
  Serial.println(pos);
  moveStepper.moveTo(p.sortedPosition);  
    while (moveStepper.distanceToGo() != 0) {
        moveStepper.run(); //    
      }
  Place();
}
void GotoPencil(Pencil &p) {      //call to move to desired pencil's position
  Serial.print("GO TO PENCIL at ");
  int pos = p.originalPosition;
  Serial.println(pos);
  moveStepper.moveTo(int(pos)); //move to target position
  moveStepper.setSpeed(200);         // set initial speed
   while (moveStepper.distanceToGo() != 0) {
        moveStepper.run(); //    
      }
}
void GoTo( int pos) {
  Serial.println("GO TO POSITION");
  moveStepper.moveTo(pos); //move to target position
  moveStepper.setSpeed(200);         // set initial speed
   while (moveStepper.distanceToGo() != 0) {
        moveStepper.run(); //    
      }  
}


//--------------Common Servo Function-------------

void Grab() {
  Serial.println(" ");
  Serial.println("GRABBING PENCIL");
  grabberDefault();
  delay(500);
  folderDown();
  delay(500);
  grabberClose();
  delay(1000);
  slowMove(folderServo, liftedFolderPosition, 30);  //might be better to directly se folderLifted()
}
void Place() {
  Serial.println(" ");
  Serial.println("PLACING PENCIL");
  slowMove(folderServo, downFolderPosition, 30);
  delay(100);
  grabberDefault();
  delay(300);
  folderLifted();
  delay(300);
}
void slowMove(Servo &servoObject, int goalPosition, int interval) {
  Serial.println("_slowMove_");
  int currentPos = servoObject.read();

  //goes to the position step by step. No condition because if pos already too low the first for never runs
  for (int pos = servoObject.read(); pos <= goalPosition; pos++) {
    servoObject.write(pos);  // tell servo to go to position in variable 'pos'
    delay(interval);
  }
  for (int pos = servoObject.read(); pos >= goalPosition; pos--) {
    servoObject.write(pos);  // tell servo to go to position in variable 'pos'
    delay(interval);
  }
}

//--------------Utility Folder Functions-------------

void folderUp() {
  Serial.print("_Arm Up_");
  slowMove(folderServo, defaultFolderPosition, 30);
}
void folderDown() {
  Serial.print("_Arm Down_");
  slowMove(folderServo, downFolderPosition, 30);
}
void folderLifted() {
  Serial.print("_Arm Lifted_");
  folderServo.write(liftedFolderPosition);
}

//--------------Utility Grabber Functions-------------

void grabberOpen() {
  Serial.print("_Opening Claws_");
  grabberServo.write(openAngle);
  //slowMove(grabberServo, openAngle, 5);
}
void grabberFullyOpen() {
  Serial.print("_Opening Claws_");
  grabberServo.write(fullyOpenAngle);
  //slowMove(grabberServo, openAngle, 5);
}
void grabberClose() {
  Serial.print("_Closing Claws_");
  grabberServo.write(closedAngle);
}
void grabberDefault() {
  Serial.print("_Grabber Return To Default_");
  grabberServo.write(defaultGrabberPosition);
}
void grabberWave() {      //grabber waves to ask for user action
  slowMove(folderServo, liftedFolderPosition, 30);
  delay(random(400, 700));
  grabberDefault();
  delay(random(600, 1000));
  grabberOpen();    
  delay(random(600, 1000));
  grabberDefault();
  delay(random(600, 1000));
  grabberOpen();


}

//--------------Conversion - Calculations-------------

double distanceToStep(unsigned int distance) {
  float stepAngle = 0.088;   //movement per step in degrees
  float distancePerStep = 2 * 21 * PI * (stepAngle / 360);
  int stepNumber = int(distance / distancePerStep);
  return stepNumber;                         //stepsPerRevolution * 2 because of half-step mode
}

bool cmpfuncHue(Pencil a, Pencil b) { 
  return a.hue < b.hue; 
}
bool cmpfuncSat(Pencil a, Pencil b) { 
  return a.sat < b.sat; 
}
bool cmpfuncValue(Pencil a, Pencil b) { 
  return a.value < b.value; 
}

void sortPencils() {
  std::sort(std::begin(pencils), std::end(pencils), cmpfuncValue);
  for(int i = 0; i < penCount; i++) {
    pencils[i].sortedPosition = i * distanceToStep(20);   //sets the sorted position of the pens as i * 4cm
  }
}

//--------------Sensor Functions-------------

void sensorCalibration() {

  Serial.println("STARTING CALIBRATION");
  
  for(byte a = 0; a < 2; a++) {        //repeats twice, once for black - once for white

      //Signals the user to place the first calibration material Black or White
      grabberWave();
      int t = millis() / 1000;
      Serial.println("Waiting for user to place B or W and press button");
      do {
        bool ledState = digitalRead(LED_BUILTIN);
        buttonState = digitalRead(buttonPin);
        if ((((millis() / 1000) - t) / 2) % 1) {
          digitalWrite(LED_BUILTIN, !ledState);
        }
        delay(50); 
      } while(buttonState == LOW);

      grabberFullyOpen();
      delay(200);
      folderLifted(); //After calibrate material was places, lowers the arm

      t = millis() / 1000; //get number of iterations to run to reach calibrationDuration
      Serial.print("Calibrating");
      do {
        Serial.print(".");
        tcs.setInterrupt(false);        
        uint16_t r, g, b, c;
        tcs.getRawData(&r, &g, &b, &c);    //Start measurement

        if(r <= redLow){redLow = r;}
        else if(r >= redHigh){redHigh = r;}
        if(g <= greenLow){greenLow = g;}
        else if(g >= greenHigh){greenHigh = g;}
        if(b <= blueLow){blueLow = b;}
        else if(b >= blueHigh){blueHigh = b;} 
        delay(50);
      } while ((millis() / 1000) - t <= calibrationDuration);

  }
  folderLifted();
  grabberDefault();
  Serial.println("CALIBRATION FINISHED");
}

//sub-routine for sensing a color
Color senseColour(bool directRGB, bool sensorDown, bool sensorReturn) { 

  if(sensorDown){
    grabberFullyOpen();  //fully brings the sensor down
    delay(400);     //evens out the power draaw a bit
    folderLifted();   //makes sure the arm is down    
  }

  int avgRed, avgGreen, avgBlue;

  int sumRed = 0;
  int sumGreen = 0;
  int sumBlue = 0;
  
  for (int i = 0; i < averagingCount; i++){
    tcs.setInterrupt(true);      //Start measurement
    delay(60);                    //Takes 50ms to read
    
    uint16_t r, g, b, c;
      tcs.getRawData(&r, &g, &b, &c);
        Serial.println("Remapping");
        r = map(r, redLow, redHigh, 0, 255);
        r = constrain(r, 0, 255);
        sumRed += r;
      
        g = map(g, greenLow, greenHigh, 0, 255);
        g = constrain(g, 0, 255);
        sumGreen += g;
      
        b = map(b, blueLow, blueHigh, 0, 255);
        b = constrain(b, 0, 255);
        sumBlue += b;
       
  }
  
  avgRed = sumRed/averagingCount;  avgGreen = sumGreen/averagingCount;  avgBlue = sumBlue/averagingCount;

  Serial.println(" ");
  Serial.print("avgRed: ");Serial.print(avgRed);
  Serial.print(", ");
  Serial.print("avgGreen: ");Serial.print(avgGreen);
  Serial.print(", ");
  Serial.print("avgBlue: ");Serial.println(avgBlue);

  Color col(avgRed, avgGreen, avgBlue);
  
  if(sensorReturn){grabberServo.write(defaultGrabberPosition);}       // sets the sensor back up
  return col;
}

float euclidianDistance(Color c1, Color c2) {
  //Computes single value difference between two colors
  //Output range [0.0,3.0] because it skips the square root
  float redDifference = (c1.red / 255.00) - (c2.red / 255.00);
  //Serial.print("redDifference"); Serial.print(redDifference);
  float greenDifference = (c1.green / 255.00) - (c2.green / 255.00);
  //Serial.print("greenDifference"); Serial.print(greenDifference);
  float blueDifference = (c1.blue / 255.00) - (c2.blue / 255.00);
  //Serial.print("blueDifference"); Serial.print(blueDifference);
  float eD = redDifference * redDifference + greenDifference * greenDifference + blueDifference * blueDifference; //sqrt not done for faster computing
  return eD;
}
